
 --
 -- 24-to-8-bit
 --
 -- Created by Rod Jackson
 -- Copyright August, 2001
 -- All rights reserved
 --
 -- Syntax (unbound): ex 24to8bit inbmp outbmp
 -- Syntax (bound):   24to8bit inbmp outbmp
 --
 -- This program is to be run from the command line. It
 -- reads the specified 24-bit color bitmap file and
 -- dithers it down to 8-bit color. It then writes the
 -- modified bitmap out to the second specified file.
 --
 -- NOTE: If no filename extension is given,
 -- ".BMP" is assumed.
 --

 --------------
 -- INCLUDES --
 --------------

include bmp_data.e
include newimage.e

 ---------------
 -- CONSTANTS --
 ---------------

constant EMPTY_SEQ = {}

constant BMP_EXTENSION = ".bmp"

constant PROGRAM_ERROR_CODE = 24 / 8 -- 3

constant BLANK_COLOR = {0,0,0}
constant WHITE_COLOR = {63,63,63}

 ----------
 -- VARS --
 ----------

integer  bmp_w, bmp_h, highest_index, highest_matches, best_index,
	 depth, num_colors, curr_num_matches, err
atom     prior_comp, new_comp, old_color_scaling1, old_color_scaling2,
         old_color_scaling3
sequence bmp_data, bmp_pal, bmp_image, cmd_line, in_bmp_file,
	 out_bmp_file, image_slice, new_pal, matches, scalings,
	 old_color_scaling, prior_comparison, new_comparison,
         scalingsj
object   curr_color

new_comparison = {0,0,0}


 --------------
 -- ROUTINES --
 --------------

function abs (atom x)
   if (x < 0) then
      return -x
   else
      return x
   end if
   --return ((x >= 0.0) - (x < 0.0)) * x
end function

 ----------

 -- borrowed & modified from Jeffrey Fielding's neural net routines
function scale (object value, object min, object max)

   return (value - min) / (max - min)

end function -- scale()

 ----------

atom accum
function sum (sequence s)
   accum = 0
   for i = 1 to length (s) do
      accum = accum + s[i]
   end for
   return accum
end function


 ---------------
 -- MAIN CODE --
 ---------------

procedure main ()

 -- extract command-line info (if none, offer help then quit)
cmd_line = command_line ()

if (length (cmd_line) < 4) then
   puts (1, "\n" &
      "***  24-to-8-bit v. 1.0 - Bitmap Conversion  ***\n" &
      "        Copyright Aug. 2001 by Rod Jackson\n" &
      "- for more information: rodjackson@bigfoot.com -\n" &
      "------------------------------------------------\n\n" &
      "SYNTAX:  24to8bit inbmp outbmp\n\n" &
      "NOTE!! Only 24-bit bitmaps will be altered; bitmaps\n" &
      "   of lower color depth will be written as is.\n\n")
   abort (0)
end if

 -- extract filenames (assume a .BMP extension)
in_bmp_file  = cmd_line[3]
out_bmp_file = cmd_line[4]
if (find ('.', in_bmp_file) < 1) then
   in_bmp_file = in_bmp_file & BMP_EXTENSION
end if
if (find ('.', out_bmp_file) < 1) then
   out_bmp_file = out_bmp_file & BMP_EXTENSION
end if

 -- save a wee bit of space...
cmd_line = EMPTY_SEQ

 -- get the color depth, so we know which routines
 -- to use and whether or not to actually modify
depth = get_color_depth (in_bmp_file)

puts (1,"Loading bitmap")

 -- now load the actual bitmap. Be sure to save space!
if (depth = 24) then
   bmp_data = get_24bit_bmp_data (in_bmp_file)
else
   bmp_data = get_bmp_data (in_bmp_file)
   bmp_data[1] = bmp_data[2]
   bmp_data[2] = bmp_data[3]
end if
bmp_w = bmp_data[1]
bmp_h = bmp_data[2]
if (depth = 24) then
   bmp_data = read_24bit_bitmap (in_bmp_file)
else
   bmp_data = read_bitmap (in_bmp_file)
end if
bmp_pal = bmp_data[1]
bmp_image = bmp_data[2]

puts (1, "Establishing palette...\n")

 -- we've got to come up with a standard number of colors
num_colors = length (bmp_pal)
if (num_colors < 256) then
    -- fill palette to meet an 8, 4, or 2-bit color depth
   if (num_colors > 16) then
      for i = 1 to 256 - num_colors do
         bmp_pal = append (bmp_pal, BLANK_COLOR)
      end for
   elsif (num_colors > 4) and (num_colors < 16) then
      for i = 1 to 16 - num_colors do
	 bmp_pal = append (bmp_pal, BLANK_COLOR)
      end for
   elsif (num_colors = 3) then
      bmp_pal = append (bmp_pal, BLANK_COLOR)
   end if
       
elsif (num_colors > 256) then
   -- NOW we have to really work; we need to shrink the palette
   -- so it fits within an 8-bit depth (256 colors)

   puts (1, "Fitting large palette into 256 colors...\n")

    -- first, pick the top 256 colors into a new palette
   matches = bmp_data[3]
   bmp_data = EMPTY_SEQ
   new_pal = repeat (0, 256)
   for i = 1 to 256 do
      highest_index = 0
      highest_matches = 0
      for j = 1 to num_colors do
	 curr_num_matches = matches[j]
	 if (curr_num_matches > highest_matches) then
	    highest_index = j
	    highest_matches = curr_num_matches
	 end if
      end for
      new_pal[i] = bmp_pal[highest_index]
      bmp_pal[highest_index] = i
      matches[highest_index] = 0
   end for

printf (1, "Mapping %d colors into top 256", {num_colors})

    -- next, map the unchosen colors into the new palette
    -- begin by scaling the new colors for comparison
   scalings = repeat (0, 256)
   for i = 1 to 256 do
      scalings[i] = scale (new_pal[i], BLANK_COLOR, WHITE_COLOR)
   end for
   for i = 1 to num_colors do
      curr_color = bmp_pal[i]
      puts (1, '.')
      if (sequence (curr_color)) then
	  -- find the new palette color that scales closest
	 best_index = 1
	 old_color_scaling = scale (curr_color, BLANK_COLOR, WHITE_COLOR)
         prior_comparison = old_color_scaling - scalings[1]
         prior_comp = power (prior_comparison[1], 2) +
                      power (prior_comparison[2], 2) +
                      power (prior_comparison[3], 2)
         old_color_scaling1 = old_color_scaling[1]
         old_color_scaling2 = old_color_scaling[2]
         old_color_scaling3 = old_color_scaling[3]
	 for j = 2 to 256 do
            --new_comparison = old_color_scaling - scalings[j]
            scalingsj = scalings[j]
            --new_comparison[1] = old_color_scaling[1] - scalingsj[1]
            --new_comparison[2] = old_color_scaling[2] - scalingsj[2]
            --new_comparison[3] = old_color_scaling[3] - scalingsj[3]
            --new_comp = power (new_comparison[1], 2) +
            --           power (new_comparison[2], 2) +
            --           power (new_comparison[3], 2)
            new_comp = power (old_color_scaling1 - scalingsj[1], 2) +
                       power (old_color_scaling2 - scalingsj[2], 2) +
                       power (old_color_scaling3 - scalingsj[3], 2)
	    if new_comp < prior_comp then
	       best_index = j
	       prior_comp = new_comp
	    end if
	 end for
	  -- we've determined the best color match
	 bmp_pal[i] = best_index
      end if
   end for
   puts (1, '\n')

puts (1, "Remapping image colors to new palette...\n")

    -- finally, map the colors for the image to the new colors
   for h = 1 to bmp_h do
      image_slice = bmp_image[h]
      for w = 1 to bmp_w do
	 image_slice[w] = bmp_pal[image_slice[w]+1] - 1
      end for
      bmp_image[h] = image_slice
   end for

    -- (and then dump the old palette for the new)
   bmp_pal = new_pal

end if -- check for number of colors in the palette

puts (1, "Writing new bitmap file...\n")

 -- write the bitmap back out now
if (depth = 24) then
   bmp_pal = bmp_pal * 4
end if
err = save_bitmap ({bmp_pal, bmp_image}, out_bmp_file)
if (err != 0) then
   puts (1, "There was an error in writing new bitmap; " &
	    "new file may not be valid.\n\n")
   abort (PROGRAM_ERROR_CODE)
end if

puts (1, "Done!\n")

end procedure -- main()

 ----------
 -- EXEC --
 ----------

main ()

